from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from dateutil.parser import parse
from dateutil.relativedelta import relativedelta

from openpyxl import load_workbook
from openpyxl.utils import get_column_letter

from werkzeug.exceptions import BadRequest

from chellow.models import Session
from chellow.utils import ct_datetime, parse_mpan_core, to_ct, to_utc

COPS = ("2", "3", "5", "10")

COLUMN_MAP = {
    "quarterly unmetered charge": "unmetered charge",
    "no. cop 5/10 meters": "no. cop 5 meters",
    "annual cop 5/10 dc/da rate": "annual cop 5 dc/da rate",
    "monthly cop 3 charge": "cop 3 charge",
    "cop 5/10 charge": "cop 5 charge",
    "monthly cop 5/10 charge": "cop 5 charge",
    "total": "grand total",
    "total ex vat": "grand total",
    "total in vat": "grand total 2",
    "total inc vat": "grand total 2",
    "vat @ 20 %": "vat @ 20%",
    "vat @20%": "vat @ 20%",
}
for cop in COPS:
    COLUMN_MAP[f"quarterly cop {cop} charge"] = f"cop {cop} charge"


def make_column_lookup(sheet):
    column_lookup = {}
    for cell in sheet[11]:
        title = str(cell.value).strip().lower()
        try:
            title = COLUMN_MAP[title]
        except KeyError:
            pass
        if title in column_lookup:
            title += " 2"
        column_lookup[title] = get_column_letter(cell.column)
    return column_lookup


def parse_ct_date(date_str):
    dt = parse(date_str, dayfirst=True)
    return ct_datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute)


def get_ct_date(sheet, col, row):
    cell = get_cell(sheet, col, row)
    val = cell.value
    if isinstance(val, Datetime):
        dt = val
    elif isinstance(val, str):
        dt = parse_ct_date(val)
    return to_ct(dt)


def get_start_date(sheet, col, row):
    dt = get_ct_date(sheet, col, row)
    return None if dt is None else to_utc(dt)


def get_cell(sheet, col, row):
    try:
        coordinates = f"{col}{row}"
        return sheet[coordinates]
    except IndexError:
        raise BadRequest(f"Can't find the cell {coordinates} on sheet {sheet}.")


def get_str(sheet, col, row):
    return get_cell(sheet, col, row).value.strip()


def get_dec(sheet, col, row):
    cell = get_cell(sheet, col, row)
    if cell.value is None:
        return None
    else:
        try:
            return Decimal(str(cell.value))
        except InvalidOperation as e:
            raise BadRequest(f"Problem parsing the number at {cell.coordinate}. {e}")


def get_gbp(sheet, col, row):
    gbp = get_dec(sheet, col, row)
    if gbp is None:
        return gbp
    else:
        return Decimal("0.00") + round(gbp, 2)


def get_int(sheet, col, row):
    return int(get_cell(sheet, col, row).value)


def _process_row(sess, sheet, row, issue_date, cl):
    mpan_core = parse_mpan_core(str(get_int(sheet, "B", row)))
    start_date_ct = get_ct_date(sheet, cl["bill from"], row)
    start_date = to_utc(start_date_ct)
    finish_date_ct = get_ct_date(sheet, cl["bill to"], row) + relativedelta(
        hours=23, minutes=30
    )
    finish_date = to_utc(finish_date_ct)

    elements = []
    days = (finish_date_ct - start_date_ct).days + 1

    el_titles = [("unmetered", "unmetered quant", "unmetered rate", "unmetered charge")]
    for cop in ("2", "3", "5", "10"):
        el_titles.append(
            (
                cop,
                f"no. cop {cop} meters",
                f"annual cop {cop} dc/da rate",
                f"cop {cop} charge",
            )
        )

    for cop, mpans_title, rate_title, net_title in el_titles:
        if mpans_title in cl:
            mpans = get_int(sheet, cl[mpans_title], row)
            rate = get_dec(sheet, cl[rate_title], row)
            net = get_gbp(sheet, cl[net_title], row)
            if net not in (None, 0):
                elements.append(
                    {
                        "name": "mpan",
                        "start_date": start_date,
                        "finish_date": finish_date,
                        "net": net,
                        "breakdown": {
                            "rate": {rate},
                            "days": days,
                            "mpan-days": mpans * days,
                            "cop": {cop},
                        },
                    }
                )

    for typ in ("adhoc", "regular"):
        hand_visits_title = f"no. hand held visits ({typ})"
        if hand_visits_title in cl:
            hand_visits = get_dec(sheet, cl[hand_visits_title], row)
            hand_rate = get_dec(sheet, cl[f"hand held visit ({typ}) rate"], row)
            hand_gbp = get_gbp(sheet, cl[f"hand held visit ({typ}) charge"], row)
            if hand_gbp != 0:
                elements.append(
                    {
                        "name": "activity",
                        "start_date": start_date,
                        "finish_date": finish_date,
                        "net": hand_gbp,
                        "breakdown": {
                            "rate": {hand_rate},
                            "activity-name": {"hand_held_visit"},
                            "visits": hand_visits,
                        },
                    }
                )

    annual_visits_count = get_int(sheet, cl["no. annual site visits"], row)
    annual_visits_rate = get_dec(sheet, cl["annual site visit rate"], row)
    annual_visits_gbp = get_gbp(sheet, cl["annual site visit charge"], row)
    if annual_visits_gbp != 0:
        elements.append(
            {
                "name": "activitiy",
                "start_date": start_date,
                "finish_date": finish_date,
                "net": annual_visits_gbp,
                "breakdown": {
                    "rate": {annual_visits_rate},
                    "count": annual_visits_count,
                    "activity-name": {"annual_site_visit"},
                },
            }
        )

    net = get_gbp(sheet, cl["grand total"], row)
    vat = Decimal("0.00")
    if "vat @ 20%" in cl:
        vat += get_gbp(sheet, cl["vat @ 20%"], row)
    breakdown = {
        "raw_lines": [],
        "settlement-status": ["settlement"],
        "vat": {20: {"vat": vat, "net": net}},
    }
    sum_el = sum(el["net"] for el in elements)
    if net != sum_el:
        elements.append(
            {
                "name": "activitiy",
                "start_date": start_date,
                "finish_date": finish_date,
                "net": net - sum_el,
                "breakdown": {
                    "activity-name": {"discrepancy"},
                },
            }
        )

    gross = Decimal("0.00")
    if "grand total 2" in cl:
        gross += get_gbp(sheet, cl["grand total 2"], row)

    return {
        "bill_type_code": "N",
        "kwh": Decimal(0),
        "net": net,
        "vat": vat,
        "gross": gross,
        "reads": [],
        "breakdown": breakdown,
        "account": mpan_core,
        "issue_date": issue_date,
        "start_date": start_date,
        "finish_date": finish_date,
        "mpan_core": mpan_core,
        "reference": "_".join(
            (
                start_date_ct.strftime("%Y%m%d"),
                finish_date_ct.strftime("%Y%m%d"),
                to_ct(issue_date).strftime("%Y%m%d"),
                mpan_core,
            )
        ),
        "elements": elements,
    }


class Parser:
    def __init__(self, f):
        self.book = load_workbook(f, data_only=True)
        self.sheet = self.book.worksheets[0]

        self.last_line = None
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_number = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        try:
            with Session() as sess:
                bills = []
                issue_date_str = get_str(self.sheet, "A", 7)
                issue_date = to_utc(parse_ct_date(issue_date_str[6:]))
                column_lookup = make_column_lookup(self.sheet)

                for row in range(12, len(self.sheet["A"]) + 1):
                    val = get_cell(self.sheet, "A", row).value
                    if val is None or val == "":
                        continue

                    self._set_last_line(row, val)
                    bill = _process_row(
                        sess, self.sheet, row, issue_date, column_lookup
                    )
                    bills.append(bill)
                    sess.rollback()

        except BadRequest as e:
            raise BadRequest(f"Row number: {row} {e.description}")

        return bills



from decimal import Decimal

from chellow.fake_batch_updater import run_import
from chellow.models import Contract, GContract, MarketRole, Participant
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_run_import_existing_batch_in_current_month(mocker, sess):
    vf = to_utc(ct_datetime(1996, 4, 1))
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    Contract.insert_non_core(sess, "configuration", "", {}, vf, None, {})
    g_cv_rate_script = {
        "cvs": {
            "EA": {
                1: {"applicable_at": utc_datetime(2020, 10, 3), "cv": 39.2000},
            }
        }
    }
    GContract.insert_industry(sess, "cv", "", {}, vf, None, g_cv_rate_script)
    ug_rate_script = {
        "ug_gbp_per_kwh": {"EA1": Decimal("40.1")},
    }
    GContract.insert_industry(sess, "ug", "", {}, vf, None, ug_rate_script)
    ccl_rate_script = {"ccl_gbp_per_kwh": Decimal("0.00339")}
    GContract.insert_industry(sess, "ccl", "", {}, vf, None, ccl_rate_script)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        vf,
        None,
        bank_holiday_rate_script,
    )
    charge_script = """
import chellow.gas.ccl
from chellow.gas.engine import g_rates
from chellow.utils import reduce_bill_hhs


def make_fake_bills(
        sess, log, last_month_start, last_month_finish, current_month_start,
        current_months_finish
):
    pass
"""
    g_contract_rate_script = {
        "gas_rate": 0.1,
        "standing_rate": 0.1,
    }
    g_contract = GContract.insert(
        sess,
        False,
        "Fusion 2020",
        charge_script,
        {},
        utc_datetime(2000, 1, 1),
        None,
        g_contract_rate_script,
    )
    g_contract.insert_g_batch(sess, f"fake_g_batch_{g_contract.id}", "Fake Batch", vf)

    sess.commit()
    messages = []

    def log(message):
        messages.append(message)

    def set_progress(progress):
        pass

    run_import(sess, log, set_progress)



from decimal import Decimal
from enum import Enum, auto
from io import BytesIO

import openpyxl

from sqlalchemy import select

from werkzeug.exceptions import BadRequest

from chellow.models import GContract, GRateScript
from chellow.rate_server import download
from chellow.utils import (
    ct_datetime,
    date_format,
    to_utc,
)


def get_rate(sheet, col, row):
    return get_decimal(sheet, col, row) / Decimal("100")


def get_decimal(sheet, col, row):
    return round(Decimal(get_value(sheet, col, row)), 4)


def normalize_text(txt):
    return None if txt is None else " ".join(txt.split()).lower()


def get_str(sheet, row, idx):
    return str(get_value(sheet, row, idx)).strip()


def get_value(sheet, col, row):
    return get_cell(sheet, col, row).value


def get_cell(sheet, col, row):
    try:
        coordinates = f"{col}{row}"
        return sheet[coordinates]
    except IndexError:
        raise BadRequest(f"Can't find the cell {coordinates} on sheet {sheet}.")


class State(Enum):
    BLANK = auto()
    NETWORK = auto()
    SYSTEM_COMMODITY = auto()
    SYSTEM_CAPACITY = auto()
    CUSTOMER_CAPACITY = auto()
    CUSTOMER_FIXED = auto()
    ADMIN_CHARGE = auto()
    EXIT_CAPACITY = auto()


def _handle_NETWORK(state, sheet, row, rates, networks):
    for cell in sheet[row][2:]:
        if cell.value is not None and len(cell.value) > 0:
            dn_text = normalize_text(cell.value)
            dn_code = DN_LOOKUP[dn_text]
            networks[dn_code] = cell.column_letter
            rates["gdn"][dn_code] = {
                "system_commodity": {"732000_and_over": {}},
                "system_capacity": {"732000_and_over": {}},
                "customer_capacity": {"732000_and_over": {}},
                "customer_fixed": {},
            }


def _handle_SECTION(state, sheet, row, rates, networks):
    b_cell = get_cell(sheet, "B", row)
    b_val = normalize_text(get_value(sheet, "B", row))
    if b_val not in (
        "ldz system commodity charges",
        "ldz system commodity charges",
        "ldz customer fixed charges",
        "ldz system capacity charges",
        "ldz customer capacity charges",
    ) and not isinstance(b_cell, openpyxl.cell.cell.MergedCell):
        section_name = SECTION_LOOKUP[state]
        rate_name = RATE_NAMES[section_name][b_val]

        for dn_code, col in networks.items():
            state_rates = rates["gdn"][dn_code][section_name]
            if rate_name == "732000_and_over":
                if state in (State.SYSTEM_CAPACITY, State.CUSTOMER_CAPACITY):
                    suffix = "_per_day"
                else:
                    suffix = ""
                over_rates = state_rates[rate_name]
                over_rates[f"gbp_per_kwh{suffix}"] = get_rate(sheet, col, row)
                over_rates["exponent"] = get_decimal(sheet, col, row + 2)
            elif rate_name == "minimum_gbp_per_kwh":
                state_rates["732000_and_over"]["minimum_gbp_per_kwh"] = get_rate(
                    sheet, col, row
                )
            elif rate_name == "minimum_gbp_per_kwh_per_day":
                state_rates["732000_and_over"]["minimum_gbp_per_kwh_per_day"] = (
                    get_rate(sheet, col, row)
                )
            else:
                state_rates[rate_name] = get_rate(sheet, col, row)


def _handle_ADMIN_CHARGE(state, sheet, row, rates, networks):
    b_val = normalize_text(get_value(sheet, "B", row))
    if b_val == "supply point admin charge":
        for dn_code, col in networks.items():
            rates["gdn"][dn_code]["cesp_administration_gbp_per_mprn"] = get_rate(
                sheet, col, row
            )


def _handle_EXIT_CAPACITY(state, sheet, row, rates, networks):
    b_val = get_str(sheet, "B", row)
    b_norm = normalize_text(b_val)
    if b_norm != "exit capacity unit rates by exit zone":
        for col in networks.values():
            v = get_value(sheet, col, row)
            if v is not None:
                rates["exit_zones"][b_val] = {
                    "exit_capacity_gbp_per_kwh_per_day": get_rate(sheet, col, row)
                }


def _handle_BLANK(state, sheet, row, rates, networks):
    pass


DN_LOOKUP = {
    "east of england": "EE",
    "london": "LO",
    "north west": "NW",
    "west midlands": "WM",
    "scotland": "SC",
    "southern": "SO",
    "northern": "NO",
    "wales & west": "WW",
}

SECTION_LOOKUP = {
    State.SYSTEM_COMMODITY: "system_commodity",
    State.SYSTEM_CAPACITY: "system_capacity",
    State.CUSTOMER_CAPACITY: "customer_capacity",
    State.CUSTOMER_FIXED: "customer_fixed",
}

HANDLER_LOOKUP = {
    State.NETWORK: _handle_NETWORK,
    State.SYSTEM_COMMODITY: _handle_SECTION,
    State.SYSTEM_CAPACITY: _handle_SECTION,
    State.CUSTOMER_CAPACITY: _handle_SECTION,
    State.CUSTOMER_FIXED: _handle_SECTION,
    State.ADMIN_CHARGE: _handle_ADMIN_CHARGE,
    State.EXIT_CAPACITY: _handle_EXIT_CAPACITY,
    State.BLANK: _handle_BLANK,
}

RATE_NAMES = {
    "system_commodity": {
        "up to 73,200 kwh per annum": "to_73200_gbp_per_kwh",
        "73,200 kwh - 732,000 kwh per annum": "73200_to_732000_gbp_per_kwh",
        "732,000 kwh per annum and above": "732000_and_over",
        "subject to a minimum rate of": "minimum_gbp_per_kwh",
    },
    "system_capacity": {
        "up to 73,200 kwh per annum": "to_73200_gbp_per_kwh_per_day",
        "73,200 kwh - 732,000 kwh per annum": "73200_to_732000_gbp_per_kwh_per_day",
        "732,000 kwh per annum and above": "732000_and_over",
        "subject to a minimum rate of": "minimum_gbp_per_kwh_per_day",
    },
    "customer_capacity": {
        "up to 73,200 kwh per annum": "to_73200_gbp_per_kwh_per_day",
        "73,200 kwh - 732,000 kwh per annum": "73200_to_732000_gbp_per_kwh_per_day",
        "732,000 kwh per annum and above": "732000_and_over",
        "subject to a minimum rate of": "minimum_gbp_per_kwh_per_day",
    },
    "customer_fixed": {
        "73,200 kwh - 732,000 kwh per annum bi annual read sites": "73200_to_732000_"
        "biannual_gbp_per_day",
        "73,200 kwh - 732,000 kwh per annum monthly read sites": "73200_to_732000_"
        "monthly_gbp_per_day",
    },
}

STATE_LOOKUP = {
    "": State.BLANK,
    None: State.BLANK,
    "network": State.NETWORK,
    "ldz system commodity charges": State.SYSTEM_COMMODITY,
    "ldz system capacity charges": State.SYSTEM_CAPACITY,
    "ldz customer capacity charges": State.CUSTOMER_CAPACITY,
    "ldz customer fixed charges": State.CUSTOMER_FIXED,
    "csep administration charge": State.ADMIN_CHARGE,
    "exit capacity unit rates by exit zone": State.EXIT_CAPACITY,
}


def tab_gdn_unit_rates(sheet, rates):
    state = State.BLANK
    networks = {}

    for row in range(1, len(sheet["B"]) + 1):
        cell_1 = get_cell(sheet, "B", row)
        if isinstance(cell_1, openpyxl.cell.cell.Cell):
            val_1 = normalize_text(get_value(sheet, "B", row))
            try:
                state = STATE_LOOKUP[val_1]
            except KeyError:
                pass

        HANDLER_LOOKUP[state](state, sheet, row, rates, networks)


def find_dn_rates(file_name, file_like):
    rates = {"a_file_name": file_name, "exit_zones": {}, "gdn": {}}

    book = openpyxl.load_workbook(file_like, data_only=True)

    for sheet in book.worksheets:
        title = sheet.title.strip().lower()
        if title.startswith("gdn unit rates"):
            tab_gdn_unit_rates(sheet, rates)

    return rates


def find_nts_1_rates(file_name, file_like):
    return find_nts_rates(file_name, file_like, "D")


def find_nts_2_rates(file_name, file_like):
    return find_nts_rates(file_name, file_like, "F")


def get_nts_rate(sheet, col, row):
    val_raw = get_value(sheet, col, row)
    return Decimal("0") if val_raw is None else Decimal(val_raw) / Decimal("100")


def find_nts_rates(file_name, file_like, col):
    rates = {"a_file_name": file_name}

    book = openpyxl.load_workbook(file_like, data_only=True)

    for sheet in book.worksheets:
        title = sheet.title.strip().lower()
        if title.startswith("nts unit rates") or title.startswith("ngt unit rates"):
            rates["so_entry_gbp_per_kwh"] = get_nts_rate(sheet, col, 11)
            rates["so_exit_gbp_per_kwh"] = get_nts_rate(sheet, col, 13)
            rates["to_entry_gbp_per_kwh"] = get_nts_rate(sheet, col, 10)
            rates["to_exit_gbp_per_kwh"] = get_nts_rate(sheet, col, 12)

    return rates


def rate_server_import(sess, log, set_progress, paths):
    log("Starting to check for new DN spreadsheets")

    year_entries = {}
    for path, url in paths:
        if len(path) == 4:
            year_str, utility, rate_type, file_name = path
            if utility == "gas" and rate_type == "dn":
                year = int(year_str)
                try:
                    fl_entries = year_entries[year]
                except KeyError:
                    fl_entries = year_entries[year] = {}

                fl_entries[file_name] = url

    for year, year_sheets in sorted(year_entries.items()):
        year_start = to_utc(ct_datetime(year, 4, 1))
        oct_start = to_utc(ct_datetime(year, 10, 1))
        nts_contract = GContract.get_industry_by_name(sess, "nts_commodity")
        if year_start < nts_contract.start_g_rate_script.start_date:
            continue
        nts_rs_1 = sess.execute(
            select(GRateScript).where(
                GRateScript.g_contract == nts_contract,
                GRateScript.start_date == year_start,
            )
        ).scalar_one_or_none()
        if nts_rs_1 is None:
            nts_rs_1 = nts_contract.insert_g_rate_script(sess, year_start, {})

        nts_rs_2 = sess.execute(
            select(GRateScript).where(
                GRateScript.g_contract == nts_contract,
                GRateScript.start_date == oct_start,
            )
        ).scalar_one_or_none()
        if nts_rs_2 is None:
            nts_rs_2 = nts_contract.insert_g_rate_script(sess, oct_start, {})

        dn_contract = GContract.get_industry_by_name(sess, "dn")
        dn_rs = sess.execute(
            select(GRateScript).where(
                GRateScript.g_contract == dn_contract,
                GRateScript.start_date == year_start,
            )
        ).scalar_one_or_none()
        if dn_rs is None:
            dn_rs = dn_contract.insert_g_rate_script(sess, year_start, {})

        if len(year_sheets) > 0:
            file_name, url = sorted(year_sheets.items())[-1]

            nts_rs_1_script = nts_rs_1.make_script()
            if nts_rs_1_script.get("a_file_name") != file_name:
                nts_rs_1.update(find_nts_1_rates(file_name, BytesIO(download(url))))
                log(f"Updated NTS rate script for {date_format(year_start)}")

            nts_rs_2_script = nts_rs_2.make_script()
            if nts_rs_2_script.get("a_file_name") != file_name:
                nts_rs_2.update(find_nts_2_rates(file_name, BytesIO(download(url))))
                log(f"Updated NTS rate script for {date_format(oct_start)}")

            dn_rs_script = dn_rs.make_script()
            if dn_rs_script.get("a_file_name") != file_name:
                dn_rs.update(find_dn_rates(file_name, BytesIO(download(url))))
                log(f"Updated DN rate script for {date_format(year_start)}")

    log("Finished DN spreadsheets")
    sess.commit()



import csv
from decimal import Decimal
from io import StringIO

from utils import match_tables

from chellow.models import (
    BillType,
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    ReadType,
    Site,
    Source,
    Tpr,
    User,
    UserRole,
    VoltageLevel,
    insert_bill_types,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.reports.report_219 import content
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_content(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    supply_id = supply.id
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    batch = imp_supplier_contract.insert_batch(sess, "a", "batch", vf)

    insert_bill_types(sess)
    bill_type = BillType.get_by_code(sess, "N")
    bill = batch.insert_bill(
        sess,
        "xxx",
        "543",
        to_utc(ct_datetime(2020, 2, 3)),
        to_utc(ct_datetime(2020, 1, 3)),
        to_utc(ct_datetime(2020, 1, 6)),
        Decimal("0"),
        Decimal("0.00"),
        Decimal("0.00"),
        Decimal("0.00"),
        bill_type,
        {},
        supply,
    )
    tpr = Tpr.insert(sess, "00001", False, True)
    read_type = ReadType.insert(sess, "A", "Actual")
    bill.insert_read(
        sess,
        tpr,
        Decimal(1),
        "kWh",
        "5457339",
        "3782897",
        to_utc(ct_datetime(2020, 1, 3)),
        Decimal("66"),
        read_type,
        to_utc(ct_datetime(2020, 1, 5)),
        Decimal("100"),
        read_type,
    )

    sess.commit()
    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch("chellow.reports.report_219.open_file", return_value=mock_file)
    year = 2020
    month = 1
    months = 1
    content(year, month, months, supply_id, user_id)
    mock_file.seek(0)
    actual = list(csv.reader(mock_file))
    expected = [
        [
            "duration_start",
            "duration_finish",
            "site_code",
            "site_name",
            "supply_id",
            "imp_mpan_core",
            "exp_mpan_core",
            "batch_reference",
            "bill_id",
            "bill_reference",
            "bill_issue_date",
            "bill_type",
            "register_read_id",
            "tpr",
            "coefficient",
            "prev_read_date",
            "prev_read_value",
            "prev_read_type",
            "pres_read_date",
            "pres_read_value",
            "pres_read_type",
        ],
        [
            "2020-01-01 00:00",
            "2020-01-31 23:30",
            "CI017",
            "Water Works",
            "1",
            "22 7867 6232 781",
            "",
            "a",
            "1",
            "543",
            "2020-02-03 00:00",
            "N",
            "1",
            "00001",
            "1",
            "2020-01-03 00:00",
            "66",
            "A",
            "2020-01-05 00:00",
            "100",
            "A",
        ],
    ]
    match_tables(expected, actual)



import csv
import sys
import threading
import traceback

from flask import g, redirect

import requests

from sqlalchemy import select
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null, or_

from werkzeug.exceptions import BadRequest

from chellow.dloads import open_file
from chellow.models import (
    Contract,
    Era,
    MtcParticipant,
    Party,
    ReportRun,
    Session,
    Source,
    Supply,
    User,
)
from chellow.utils import csv_make_val, req_checkbox, utc_datetime_now

FNAME = "ecoes_comparison"


def _parse_date(date_str):
    if len(date_str) > 0:
        if "/" in date_str:
            day, month, year = date_str.split("/")
            return f"{year}-{month}-{day} 00:00"
        else:
            return f"{date_str[:4]}-{date_str[4:6]}-{date_str[6:]} 00:00"
    else:
        return date_str


def content(user_id, show_ignored, report_run_id):
    f = report_run = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file(f"{FNAME}.csv", user, mode="w", newline="")
            report_run = ReportRun.get_by_id(sess, report_run_id)

            props = Contract.get_non_core_by_name(
                sess, "configuration"
            ).make_properties()

            ECOES_KEY = "ecoes"
            try:
                ecoes_props = props[ECOES_KEY]
            except KeyError:
                raise BadRequest(
                    f"The property {ECOES_KEY} cannot be found in the configuration "
                    f"properties."
                )

            for key in (
                "user_name",
                "password",
                "exclude_mpan_cores",
                "ignore_mpan_cores_msn",
            ):
                try:
                    ecoes_props[key]
                except KeyError:
                    raise BadRequest(
                        f"The property {key} cannot be found in the 'ecoes' section of "
                        f"the configuration properties."
                    )

            exclude_mpan_cores = ecoes_props["exclude_mpan_cores"]
            ignore_mpan_cores_msn = ecoes_props["ignore_mpan_cores_msn"]
            url_prefix = "https://www.ecoes.co.uk/"

            with requests.Session() as s:
                r = s.get(url_prefix)
                data = {
                    "Username": ecoes_props["user_name"],
                    "Password": ecoes_props["password"],
                }
                login_j = s.post(url_prefix, data=data, allow_redirects=False).json()
                if not login_j["Success"]:
                    raise BadRequest(f"Login to ECOES failed: {login_j['Messages']}")
                elif (
                    "RedirectUrl" in login_j and "SetPassword" in login_j["RedirectUrl"]
                ):
                    raise BadRequest(
                        "Login to ECOES failed, it looks like the password needs to be "
                        "changed"
                    )
                r = s.get(
                    f"{url_prefix}PortfolioAccess/ExportPortfolioMPANs?fileType=csv"
                )

            _process(
                sess,
                r.text.splitlines(True),
                exclude_mpan_cores,
                ignore_mpan_cores_msn,
                f,
                show_ignored,
                report_run,
            )
            report_run.update("finished")
            sess.commit()

    except BadRequest as e:
        msg = e.description
        if report_run is not None:
            report_run.update("interrupted")
            report_run.insert_row(sess, "", ["problem"], {"problem": msg}, {})
            sess.commit()
        if f is not None:
            f.write(msg)
        raise e
    except BaseException as e:
        msg = traceback.format_exc()
        if report_run is not None:
            report_run.update("interrupted")
            report_run.insert_row(sess, "", ["problem"], {"problem": msg}, {})
            sess.commit()
        sys.stderr.write(msg)
        if f is not None:
            f.write(msg)
        raise e
    finally:
        if f is not None:
            f.close()


def _process(
    sess,
    ecoes_lines,
    exclude_mpan_cores,
    ignore_mpan_cores_msn,
    f,
    show_ignored,
    report_run,
):
    writer = csv.writer(f, lineterminator="\n")
    now = utc_datetime_now()

    mpans = []

    for exp in (Era.imp_mpan_core, Era.exp_mpan_core):
        for (v,) in sess.execute(
            select(exp)
            .join(Supply)
            .join(Source)
            .join(Supply.dno)
            .filter(
                Party.dno_code.notin_(("88", "99")),
                Era.finish_date == null(),
                Source.code != "3rd-party",
                exp.notin_(exclude_mpan_cores),
                exp != null(),
            )
            .distinct()
            .order_by(exp)
        ):
            mpans.append(v)

    ecoes_mpans = {}

    parser = iter(csv.reader(ecoes_lines))
    next(parser)  # Skip titles

    for values in parser:
        ecoes_titles = [
            "mpan-core",
            "address-line-1",
            "address-line-2",
            "address-line-3",
            "address-line-4",
            "address-line-5",
            "address-line-6",
            "address-line-7",
            "address-line-8",
            "address-line-9",
            "post-code",
            "supplier",
            "supplier-dip",
            "registration-from",
            "mtc",
            "mtc-date",
            "llfc",
            "llfc-from",
            "pc",
            "ssc",
            "measurement-class",
            "energisation-status",
            "da",
            "dc",
            "mop",
            "mop-appoint-date",
            "gsp-group",
            "gsp-effective-from",
            "dno-name",
            "dno-dip",
            "Domestic Premises Indicator",
            "Duos Tariff Id",
            "Duos Tariff Id Effective From Date",
            "Market Segment Indicator",
            "Market Segment Indicator Effective From Date",
            "Data Service Dip Id",
            "Data Service Effective From Date",
            "Data Service Effective To Date",
            "Data Service Mpid",
            "Metering Service Dip Id",
            "Metering Service Effective From Date",
            "Metering Service Effective To Date",
            "Metering Service Mpid",
            "msn",
            "meter-install-date",
            "meter-type",
            "map-id",
        ]

        ecoes_row = dict(zip(ecoes_titles, map(str.strip, values)))
        mpan_core = ecoes_row["mpan-core"]
        mpan_spaces = " ".join(
            (
                mpan_core[:2],
                mpan_core[2:6],
                mpan_core[6:10],
                mpan_core[-3:],
            )
        )
        if mpan_spaces in exclude_mpan_cores:
            continue

        ecoes_row["mpan_spaces"] = mpan_spaces
        if mpan_spaces in ecoes_mpans:
            prev_row = ecoes_mpans[mpan_spaces]
            msn = ecoes_row["msn"]
            if len(msn) > 0:
                prev_msns = prev_row["msn"].split(",")
                prev_msns.append(msn)
                prev_row["msn"] = ", ".join(prev_msns)
        else:
            ecoes_mpans[mpan_spaces] = ecoes_row

    titles = (
        "mpan_core",
        "mpan_core_no_spaces",
        "ecoes_pc",
        "chellow_pc",
        "ecoes_mtc",
        "ecoes_mtc_date",
        "chellow_mtc",
        "ecoes_llfc",
        "ecoes_llfc_from",
        "chellow_llfc",
        "ecoes_ssc",
        "chellow_ssc",
        "ecoes_es",
        "chellow_es",
        "ecoes_supplier",
        "ecoes_supplier_registration_from",
        "chellow_supplier",
        "chellow_supplier_contract_name",
        "ecoes_dc",
        "chellow_dc",
        "ecoes_mop",
        "ecoes_mop_appoint_date",
        "chellow_mop",
        "ecoes_gsp_group",
        "ecoes_gsp_effective_from",
        "chellow_gsp_group",
        "ecoes_msn",
        "ecoes_msn_install_date",
        "chellow_msn",
        "ecoes_meter_type",
        "chellow_meter_type",
        "ignored",
        "problem",
    )
    writer.writerow(titles)

    for mpan_spaces, ecoes in sorted(ecoes_mpans.items()):
        problem = ""
        chellow_pc = ""
        chellow_mtc = ""
        chellow_llfc = ""
        chellow_llfc_desc = ""
        chellow_ssc = ""
        chellow_es = ""
        chellow_supplier = ""
        chellow_supplier_contract_name = ""
        chellow_supplier_contract_id = None
        chellow_dc = ""
        chellow_mop = ""
        chellow_gsp_group = ""
        chellow_msn = ""
        chellow_meter_type = ""
        chellow_supply_id = None
        chellow_era_id = None
        ignore = True
        ecoes_dno_code = ecoes["mpan-core"][:2]
        ecoes_dno = Party.get_dno_by_code(sess, ecoes_dno_code, now)
        ecoes_llfc_desc = ""
        chellow_site_code = ""
        chellow_site_name = ""
        diffs = []

        try:
            ecoes_es = ecoes["energisation-status"]
        except KeyError as e:
            raise e

        ecoes_disconnected = ecoes_es == ""
        current_chell = mpan_spaces in mpans

        if ecoes_disconnected and current_chell:
            problem += "Disconnected in ECOES, but current in Chellow. "
            ignore = False

        elif not ecoes_disconnected and not current_chell:
            era = sess.execute(
                select(Era)
                .filter(
                    or_(
                        Era.imp_mpan_core == mpan_spaces,
                        Era.exp_mpan_core == mpan_spaces,
                    ),
                )
                .order_by(Era.start_date.desc())
                .limit(1)
            ).scalar_one_or_none()
            if era is None:
                problem += f"In ECOES (as {ecoes_es}) but not in Chellow. "
            else:
                problem += f"In ECOES (as {ecoes_es}) but disconnected in Chellow. "
                if era.imp_mpan_core == mpan_spaces:
                    supplier_contract = era.imp_supplier_contract
                else:
                    supplier_contract = era.exp_supplier_contract
                chellow_supplier_contract_name = supplier_contract.name
                chellow_supplier_contract_id = supplier_contract.id
                chellow_supply_id = era.supply.id
                chellow_era_id = era.id
            ignore = False

        if current_chell:
            mpans.remove(mpan_spaces)
            era = (
                sess.scalars(
                    select(Era)
                    .where(
                        Era.finish_date == null(),
                        or_(
                            Era.imp_mpan_core == mpan_spaces,
                            Era.exp_mpan_core == mpan_spaces,
                        ),
                    )
                    .options(
                        joinedload(Era.supply).joinedload(Supply.gsp_group),
                        joinedload(Era.mop_contract)
                        .joinedload(Contract.party)
                        .joinedload(Party.participant),
                        joinedload(Era.dc_contract)
                        .joinedload(Contract.party)
                        .joinedload(Party.participant),
                        joinedload(Era.imp_supplier_contract)
                        .joinedload(Contract.party)
                        .joinedload(Party.participant),
                        joinedload(Era.exp_supplier_contract)
                        .joinedload(Contract.party)
                        .joinedload(Party.participant),
                        joinedload(Era.pc),
                        joinedload(Era.imp_llfc),
                        joinedload(Era.exp_llfc),
                        joinedload(Era.mtc_participant).joinedload(
                            MtcParticipant.meter_type
                        ),
                        joinedload(Era.ssc),
                        joinedload(Era.energisation_status),
                        joinedload(Era.channels),
                    )
                )
                .unique()
                .one()
            )
            chellow_supply_id = era.supply.id
            chellow_era_id = era.id
            chellow_es = era.energisation_status.code
            if ecoes_es != chellow_es:
                problem += "The energisation statuses don't match. "
                ignore = False
                diffs.append("es")

            if not (ecoes_es == "D" and chellow_es == "D"):
                if era.imp_mpan_core == mpan_spaces:
                    supplier_contract = era.imp_supplier_contract
                    llfc = era.imp_llfc
                else:
                    supplier_contract = era.exp_supplier_contract
                    llfc = era.exp_llfc

                chellow_pc = era.pc.code
                try:
                    if int(ecoes["pc"]) != int(chellow_pc):
                        problem += "The PCs don't match. "
                        ignore = False
                        diffs.append("pc")
                except ValueError:
                    problem += "Can't parse the PC. "
                    ignore = False

                chellow_mtc = era.mtc_participant.mtc.code
                try:
                    if int(ecoes["mtc"]) != int(chellow_mtc):
                        problem += "The MTCs don't match. "
                        ignore = False
                        diffs.append("mtc")
                except ValueError:
                    problem += "Can't parse the MTC. "
                    ignore = False

                chellow_llfc = llfc.code
                chellow_llfc_desc = llfc.description
                ecoes_llfc_code = ecoes["llfc"].zfill(3)
                if ecoes_llfc_code != chellow_llfc:
                    problem += "The LLFCs don't match. "
                    ignore = False
                    diffs.append("llfc")

                ecoes_llfc = ecoes_dno.find_llfc_by_code(sess, ecoes_llfc_code, now)
                if ecoes_llfc is not None:
                    ecoes_llfc_desc = ecoes_llfc.description

                chellow_ssc = era.ssc
                if chellow_ssc is None:
                    chellow_ssc = ""
                    chellow_ssc_int = None
                else:
                    chellow_ssc = chellow_ssc.code
                    chellow_ssc_int = int(chellow_ssc)

                if len(ecoes["ssc"]) > 0:
                    ecoes_ssc_int = int(ecoes["ssc"])
                else:
                    ecoes_ssc_int = None

                if ecoes_ssc_int != chellow_ssc_int and not (
                    ecoes_ssc_int is None and chellow_ssc_int is None
                ):
                    problem += "The SSCs don't match. "
                    ignore = False
                    diffs.append("ssc")

                chellow_supplier = supplier_contract.party.participant.code
                chellow_supplier_contract_name = supplier_contract.name
                chellow_supplier_contract_id = supplier_contract.id
                if chellow_supplier != ecoes["supplier"]:
                    problem += "The supplier codes don't match. "
                    ignore = False
                    diffs.append("supplier")

                dc_contract = era.dc_contract
                chellow_dc_props = dc_contract.make_properties()
                if "nhh_participant" in chellow_dc_props:
                    chellow_dc = chellow_dc_props["nhh_participant"]
                else:
                    chellow_dc = dc_contract.party.participant.code

                if chellow_dc != ecoes["dc"]:
                    problem += "The DC codes don't match. "
                    ignore = False
                    diffs.append("dc")

                mop_contract = era.mop_contract
                if mop_contract is None:
                    chellow_mop = ""
                else:
                    chellow_mop = mop_contract.party.participant.code

                if chellow_mop != ecoes["mop"]:
                    problem += "The MOP codes don't match. "
                    ignore = False
                    diffs.append("mop")

                chellow_gsp_group = era.supply.gsp_group.code
                if chellow_gsp_group != ecoes["gsp-group"]:
                    problem += "The GSP group codes don't match. "
                    ignore = False
                    diffs.append("gsp_group")

                chellow_site = era.get_physical_site(sess)
                chellow_site_code = chellow_site.code
                chellow_site_name = chellow_site.name

                chellow_msn = era.msn

                if set([m.strip() for m in chellow_msn.split(",")]) != set(
                    [m.strip() for m in ecoes["msn"].split(",")]
                ):
                    problem += "The meter serial numbers don't match. "
                    diffs.append("msn")
                    if mpan_spaces not in ignore_mpan_cores_msn:
                        ignore = False
                elif mpan_spaces in ignore_mpan_cores_msn:
                    problem += (
                        "This MPAN core is in mpan_cores_ignore and yet the meter "
                        "serial numbers do match. "
                    )

                chellow_dtc_meter_type = era.dtc_meter_type
                if chellow_dtc_meter_type is None:
                    chellow_meter_type = ""
                else:
                    chellow_meter_type = chellow_dtc_meter_type.code

                if chellow_meter_type != ecoes["meter-type"]:
                    problem += "The DTC meter types don't match."
                    ignore = False
                    diffs.append("meter_type")

        if len(problem) > 0 and not (not show_ignored and ignore):
            address_lines = [ecoes[f"address-line-{i}"] for i in range(1, 10)]
            values = {
                "mpan_core": mpan_spaces,
                "mpan_core_no_spaces": ecoes["mpan-core"],
                "ecoes_pc": ecoes["pc"],
                "chellow_pc": chellow_pc,
                "ecoes_mtc": ecoes["mtc"],
                "ecoes_mtc_date": _parse_date(ecoes["mtc-date"]),
                "chellow_mtc": chellow_mtc,
                "ecoes_llfc": ecoes["llfc"],
                "ecoes_llfc_desc": ecoes_llfc_desc,
                "ecoes_llfc_from": _parse_date(ecoes["llfc-from"]),
                "chellow_llfc": chellow_llfc,
                "chellow_llfc_desc": chellow_llfc_desc,
                "ecoes_ssc": ecoes["ssc"],
                "chellow_ssc": chellow_ssc,
                "ecoes_es": ecoes["energisation-status"],
                "chellow_es": chellow_es,
                "ecoes_supplier": ecoes["supplier"],
                "ecoes_supplier_registration_from": _parse_date(
                    ecoes["registration-from"]
                ),
                "chellow_supplier": chellow_supplier,
                "chellow_supplier_contract_name": chellow_supplier_contract_name,
                "ecoes_dc": ecoes["dc"],
                "chellow_dc": chellow_dc,
                "ecoes_mop": ecoes["mop"],
                "ecoes_mop_appoint_date": _parse_date(ecoes["mop-appoint-date"]),
                "chellow_mop": chellow_mop,
                "ecoes_gsp_group": ecoes["gsp-group"],
                "ecoes_gsp_effective_from": _parse_date(ecoes["gsp-effective-from"]),
                "chellow_gsp_group": chellow_gsp_group,
                "ecoes_msn": ecoes["msn"],
                "chellow_msn": chellow_msn,
                "ecoes_msn_install_date": _parse_date(ecoes["meter-install-date"]),
                "ecoes_meter_type": ecoes["meter-type"],
                "chellow_meter_type": chellow_meter_type,
                "ignored": ignore,
                "problem": problem,
                "address": ", ".join(
                    [a for a in address_lines if len(a) > 0 and a != "NULL"]
                ),
                "chellow_site_code": chellow_site_code,
                "chellow_site_name": chellow_site_name,
            }
            writer.writerow(csv_make_val(values[t]) for t in titles)
            values["chellow_supplier_contract_id"] = chellow_supplier_contract_id
            values["chellow_supply_id"] = chellow_supply_id
            values["diffs"] = diffs
            values["chellow_era_id"] = chellow_era_id
            report_run.insert_row(sess, "", titles, values, {})
            sess.commit()

    for mpan_core in mpans:
        supply = Supply.get_by_mpan_core(sess, mpan_core)
        era = supply.find_era_at(sess, None)
        if era.imp_mpan_core == mpan_core:
            supplier_contract = era.imp_supplier_contract
            llfc = era.imp_llfc
        else:
            supplier_contract = era.exp_supplier_contract
            llfc = era.exp_llfc

        ssc = "" if era.ssc is None else era.ssc.code

        es = era.energisation_status.code

        dc_contract = era.dc_contract
        dc = "" if dc_contract is None else dc_contract.party.participant.code

        mop_contract = era.mop_contract
        mop = "" if mop_contract is None else mop_contract.party.participant.code

        msn = "" if era.msn is None else era.msn

        dtc_meter_type = era.dtc_meter_type
        meter_type = "" if dtc_meter_type is None else dtc_meter_type.code

        values = {
            "mpan_core": mpan_core,
            "mpan_core_no_spaces": mpan_core.replace(" ", ""),
            "ecoes_pc": "",
            "chellow_pc": era.pc.code,
            "ecoes_mtc": "",
            "ecoes_mtc_date": "",
            "chellow_mtc": era.mtc_participant.mtc.code,
            "ecoes_llfc": "",
            "ecoes_llfc_from": "",
            "chellow_llfc": llfc.code,
            "ecoes_ssc": "",
            "chellow_ssc": ssc,
            "ecoes_es": "",
            "chellow_es": es,
            "ecoes_supplier": "",
            "ecoes_supplier_registration_from": "",
            "chellow_supplier": supplier_contract.party.participant.code,
            "chellow_supplier_contract_name": supplier_contract.name,
            "ecoes_dc": "",
            "chellow_dc": dc,
            "ecoes_mop": "",
            "ecoes_mop_appoint_date": "",
            "chellow_mop": mop,
            "ecoes_gsp_group": "",
            "ecoes_gsp_effective_from": "",
            "chellow_gsp_group": supply.gsp_group.code,
            "ecoes_msn": "",
            "chellow_msn": msn,
            "ecoes_msn_install_date": "",
            "ecoes_meter_type": "",
            "chellow_meter_type": meter_type,
            "ignored": False,
            "problem": "In Chellow, but not in ECOES.",
        }
        writer.writerow(csv_make_val(values[t]) for t in titles)
        values["chellow_supplier_contract_id"] = supplier_contract.id
        values["chellow_supply_id"] = era.supply.id
        values["diffs"] = []
        values["chellow_era_id"] = era.id
        report_run.insert_row(sess, "", titles, values, {})


def do_get(sess):
    show_ignored = req_checkbox("show_ignored")
    report_run = ReportRun.insert(
        sess,
        FNAME,
        g.user,
        FNAME,
        {},
    )
    sess.commit()
    args = g.user.id, show_ignored, report_run.id
    threading.Thread(target=content, args=args).start()
    return redirect(f"/report_runs/{report_run.id}", 303)



import csv
import sys
import threading
import traceback


from flask import g, redirect

from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true

from werkzeug.exceptions import BadRequest

from chellow.dloads import open_file
from chellow.e.computer import contract_func, forecast_date
from chellow.gas.engine import GDataSource
from chellow.models import GContract, GEra, Session, Site, SiteGEra, User
from chellow.utils import (
    c_months_u,
    date_format,
    hh_max,
    hh_min,
    make_val,
    req_date,
    req_int,
    to_ct,
)


def content(start_date, finish_date, g_contract_id, user_id):
    report_context = {}
    f = writer = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file("gas_virtual_bills.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")

            g_contract = GContract.get_by_id(sess, g_contract_id)
            forecast_dt = forecast_date()

            start_date_ct, finish_date_ct = to_ct(start_date), to_ct(finish_date)

            month_pairs = c_months_u(
                start_year=start_date_ct.year,
                start_month=start_date_ct.month,
                finish_year=finish_date_ct.year,
                finish_month=finish_date_ct.month,
            )

            bill_titles = contract_func(
                report_context, g_contract, "virtual_bill_titles"
            )()
            writer.writerow(
                ["MPRN", "Site Code", "Site Name", "Account", "From", "To"]
                + bill_titles
            )

            for month_start, month_finish in month_pairs:
                period_start = hh_max(start_date, month_start)
                period_finish = hh_min(finish_date, month_finish)

                for g_era in (
                    sess.query(GEra)
                    .distinct()
                    .filter(
                        GEra.g_contract == g_contract,
                        GEra.start_date <= period_finish,
                        or_(
                            GEra.finish_date == null(), GEra.finish_date >= period_start
                        ),
                    )
                ):
                    chunk_start = hh_max(g_era.start_date, period_start)
                    chunk_finish = hh_min(g_era.finish_date, period_finish)

                    data_source = GDataSource(
                        sess,
                        chunk_start,
                        chunk_finish,
                        forecast_dt,
                        g_era,
                        report_context,
                        None,
                    )

                    site = (
                        sess.query(Site)
                        .join(SiteGEra)
                        .filter(SiteGEra.g_era == g_era, SiteGEra.is_physical == true())
                        .one()
                    )

                    vals = [
                        data_source.mprn,
                        site.code,
                        site.name,
                        data_source.account,
                        date_format(data_source.start_date),
                        date_format(data_source.finish_date),
                    ]

                    contract_func(report_context, g_contract, "virtual_bill")(
                        data_source
                    )
                    bill = data_source.bill
                    for title in bill_titles:
                        if title in bill:
                            val = make_val(bill[title])
                            del bill[title]
                        else:
                            val = ""
                        vals.append(val)

                    for k in sorted(bill.keys()):
                        vals.append(k)
                        vals.append(str(bill[k]))
                    writer.writerow(vals)

    except BadRequest as e:
        writer.writerow(["Problem: " + e.description])
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        if writer is not None:
            writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    start_date = req_date("start")
    finish_date = req_date("finish")
    g_contract_id = req_int("g_contract_id")

    args = start_date, finish_date, g_contract_id, g.user.id
    threading.Thread(target=content, args=args).start()
    return redirect("/downloads", 303)



from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from openpyxl import load_workbook

from werkzeug.exceptions import BadRequest

from chellow.utils import parse_mpan_core, to_ct, to_utc, validate_hh_start

# "Time stamp","Value","Events","Comment","User"
# "18/10/2024 09:30:00","88.0","","",""
# "18/10/2024 09:00:00","89.1","","",""


def create_parser(reader, mpan_map, messages):
    return HhParserSchneiderXlsx(reader, mpan_map, messages)


def get_cell(sheet, col, row):
    try:
        coordinates = f"{col}{row}"
        return sheet[coordinates]
    except IndexError:
        raise BadRequest(f"Can't find the cell {coordinates} on sheet {sheet}.")


def get_date(sheet, col, row):
    cell = get_cell(sheet, col, row)
    val = cell.value
    if not isinstance(val, Datetime):
        raise BadRequest(
            f"Problem reading {val} (of type {type(val)}) as a timestamp at "
            f"{cell.coordinate}."
        )
    return val


def get_str(sheet, col, row):
    return get_cell(sheet, col, row).value.strip()


def get_dec(sheet, col, row):
    return get_dec_from_cell(get_cell(sheet, col, row))


def get_dec_from_cell(cell):
    try:
        return Decimal(str(cell.value))
    except InvalidOperation as e:
        raise BadRequest(f"Problem parsing the number at {cell.coordinate}. {e}")


def get_int(sheet, col, row):
    return int(get_cell(sheet, col, row).value)


def find_hhs(sheet, set_line_number, mpan_core):
    pres_reading = None
    for row in range(2, len(sheet["A"]) + 1):
        set_line_number(row)
        try:
            timestamp_cell = get_cell(sheet, "A", row)
            timestamp_str = timestamp_cell.value
            if timestamp_str is None:
                continue

            # 2024-10-02 00:00:00 +1H, DST
            ts_str = timestamp_str.split(",")[0]
            ts_naive = Datetime.strptime(ts_str[:19], "%Y-%m-%d %H:%M:%S")
            # Sometimes has seconds
            ts_naive = Datetime(
                ts_naive.year,
                ts_naive.month,
                ts_naive.day,
                ts_naive.hour,
                ts_naive.minute,
            )
            is_dst = ts_str[21] == 1
            if is_dst:
                start_date = to_utc(to_ct(ts_naive))
            else:
                start_date = to_utc(ts_naive)
            start_date = validate_hh_start(start_date)

            reading = get_dec(sheet, "B", row)

            if pres_reading is not None:
                value = pres_reading - reading
                if value >= 0:
                    yield {
                        "mpan_core": mpan_core,
                        "channel_type": "ACTIVE",
                        "start_date": start_date,
                        "value": value,
                        "status": "A",
                    }
            pres_reading = reading
        except BadRequest as e:
            e.description = f"Problem at line number: {row}: {e.description}"
            raise e


class HhParserSchneiderXlsx:
    def __init__(self, input_stream, mpan_map, messages):
        book = load_workbook(input_stream, data_only=True)
        self.sheet = book.worksheets[0]
        imp_name = get_str(self.sheet, "B", 1).strip()
        self.mpan_core = parse_mpan_core(mpan_map[imp_name])
        self.line_number = 0

    def _set_line_number(self, line_number):
        self.line_number = line_number

    def __iter__(self):
        return find_hhs(self.sheet, self._set_line_number, self.mpan_core)

    def close(self):
        self.shredder.close()
